@@ -1,151 +1,123 @@
+from __future__ import annotations
+
 import abc
 import copy
+import dataclasses
+import math
 import re
 import string
+import sys
 
 from datetime import date
 from datetime import datetime
 from datetime import time
 from datetime import tzinfo
 from enum import Enum
-from functools import lru_cache
 from typing import TYPE_CHECKING
 from typing import Any
+from typing import Callable
 from typing import Collection
-from typing import Dict
 from typing import Iterable
 from typing import Iterator
-from typing import List
-from typing import Optional
 from typing import Sequence
 from typing import TypeVar
-from typing import Union
 from typing import cast
 from typing import overload
 
 from tomlkit._compat import PY38
 from tomlkit._compat import decode
+from tomlkit._types import _CustomDict
+from tomlkit._types import _CustomFloat
+from tomlkit._types import _CustomInt
+from tomlkit._types import _CustomList
+from tomlkit._types import wrap_method
 from tomlkit._utils import CONTROL_CHARS
 from tomlkit._utils import escape_string
 from tomlkit.exceptions import InvalidStringError
 
 
-if TYPE_CHECKING:  # pragma: no cover
-    # Define _CustomList and _CustomDict as a workaround for:
-    # https://github.com/python/mypy/issues/11427
-    #
-    # According to this issue, the typeshed contains a "lie"
-    # (it adds MutableSequence to the ancestry of list and MutableMapping to
-    # the ancestry of dict) which completely messes with the type inference for
-    # Table, InlineTable, Array and Container.
-    #
-    # Importing from builtins is preferred over simple assignment, see issues:
-    # https://github.com/python/mypy/issues/8715
-    # https://github.com/python/mypy/issues/10068
-    from builtins import dict as _CustomDict  # noqa: N812, TC004
-    from builtins import list as _CustomList  # noqa: N812, TC004
-
-    # Allow type annotations but break circular imports
+if TYPE_CHECKING:
     from tomlkit import container
-else:
-    from collections.abc import MutableMapping
-    from collections.abc import MutableSequence
-
-    class _CustomList(MutableSequence, list):
-        """Adds MutableSequence mixin while pretending to be a builtin list"""
-
-    class _CustomDict(MutableMapping, dict):
-        """Adds MutableMapping mixin while pretending to be a builtin dict"""
 
 
 ItemT = TypeVar("ItemT", bound="Item")
+Encoder = Callable[[Any], "Item"]
+CUSTOM_ENCODERS: list[Encoder] = []
+AT = TypeVar("AT", bound="AbstractTable")
+
+
+class _ConvertError(TypeError, ValueError):
+    """An internal error raised when item() fails to convert a value.
+    It should be a TypeError, but due to historical reasons
+    it needs to subclass ValueError as well.
+    """
 
 
 @overload
-def item(
-    value: bool, _parent: Optional["Item"] = ..., _sort_keys: bool = ...
-) -> "Bool":
+def item(value: bool, _parent: Item | None = ..., _sort_keys: bool = ...) -> Bool:
+    ...
+
+
+@overload
+def item(value: int, _parent: Item | None = ..., _sort_keys: bool = ...) -> Integer:
+    ...
+
+
+@overload
+def item(value: float, _parent: Item | None = ..., _sort_keys: bool = ...) -> Float:
+    ...
+
+
+@overload
+def item(value: str, _parent: Item | None = ..., _sort_keys: bool = ...) -> String:
     ...
 
 
 @overload
 def item(
-    value: int, _parent: Optional["Item"] = ..., _sort_keys: bool = ...
-) -> "Integer":
+    value: datetime, _parent: Item | None = ..., _sort_keys: bool = ...
+) -> DateTime:
+    ...
+
+
+@overload
+def item(value: date, _parent: Item | None = ..., _sort_keys: bool = ...) -> Date:
+    ...
+
+
+@overload
+def item(value: time, _parent: Item | None = ..., _sort_keys: bool = ...) -> Time:
     ...
 
 
 @overload
 def item(
-    value: float, _parent: Optional["Item"] = ..., _sort_keys: bool = ...
-) -> "Float":
+    value: Sequence[dict], _parent: Item | None = ..., _sort_keys: bool = ...
+) -> AoT:
     ...
 
 
 @overload
-def item(
-    value: str, _parent: Optional["Item"] = ..., _sort_keys: bool = ...
-) -> "String":
+def item(value: Sequence, _parent: Item | None = ..., _sort_keys: bool = ...) -> Array:
     ...
 
 
 @overload
-def item(
-    value: datetime, _parent: Optional["Item"] = ..., _sort_keys: bool = ...
-) -> "DateTime":
+def item(value: dict, _parent: Array = ..., _sort_keys: bool = ...) -> InlineTable:
     ...
 
 
 @overload
-def item(
-    value: date, _parent: Optional["Item"] = ..., _sort_keys: bool = ...
-) -> "Date":
+def item(value: dict, _parent: Item | None = ..., _sort_keys: bool = ...) -> Table:
     ...
 
 
 @overload
-def item(
-    value: time, _parent: Optional["Item"] = ..., _sort_keys: bool = ...
-) -> "Time":
+def item(value: ItemT, _parent: Item | None = ..., _sort_keys: bool = ...) -> ItemT:
     ...
 
 
-@overload
-def item(
-    value: Sequence[dict], _parent: Optional["Item"] = ..., _sort_keys: bool = ...
-) -> "AoT":
-    ...
-
-
-@overload
-def item(
-    value: Sequence, _parent: Optional["Item"] = ..., _sort_keys: bool = ...
-) -> "Array":
-    ...
-
-
-@overload
-def item(value: dict, _parent: "Array" = ..., _sort_keys: bool = ...) -> "InlineTable":
-    ...
-
-
-@overload
-def item(
-    value: dict, _parent: Optional["Item"] = ..., _sort_keys: bool = ...
-) -> "Table":
-    ...
-
-
-@overload
-def item(
-    value: ItemT, _parent: Optional["Item"] = ..., _sort_keys: bool = ...
-) -> ItemT:
-    ...
-
-
-def item(
-    value: Any, _parent: Optional["Item"] = None, _sort_keys: bool = False
-) -> "Item":
+def item(value: Any, _parent: Item | None = None, _sort_keys: bool = False) -> Item:
     """Create a TOML item from a Python object.
 
     :Example:
@@ -171,21 +143,23 @@ def item(
     elif isinstance(value, float):
         return Float(value, Trivia(), str(value))
     elif isinstance(value, dict):
-        table_constructor = InlineTable if isinstance(_parent, Array) else Table
+        table_constructor = (
+            InlineTable if isinstance(_parent, (Array, InlineTable)) else Table
+        )
         val = table_constructor(Container(), Trivia(), False)
         for k, v in sorted(
             value.items(),
-            key=lambda i: (isinstance(i[1], dict), i[0] if _sort_keys else 1),
+            key=lambda i: (isinstance(i[1], dict), i[0]) if _sort_keys else 1,
         ):
             val[k] = item(v, _parent=val, _sort_keys=_sort_keys)
-        only_child = val[next(iter(value))] if len(value) == 1 else None
-        if table_constructor is Table and isinstance(only_child, (AoT, Table)):
-            # The table becomes super table if the only child is a table or AoT.
-            val._is_super_table = True
 
         return val
     elif isinstance(value, (list, tuple)):
-        if value and all(isinstance(v, dict) for v in value):
+        if (
+            value
+            and all(isinstance(v, dict) for v in value)
+            and (_parent is None or isinstance(_parent, Table))
+        ):
             a = AoT([])
             table_constructor = Table
         else:
@@ -200,7 +174,7 @@ def item(
                     v.items(),
                     key=lambda i: (isinstance(i[1], dict), i[0] if _sort_keys else 1),
                 ):
-                    i = item(_v, _parent=a, _sort_keys=_sort_keys)
+                    i = item(_v, _parent=table, _sort_keys=_sort_keys)
                     if isinstance(table, InlineTable):
                         i.trivia.trail = ""
 
@@ -238,14 +212,20 @@ def item(
             Trivia(),
             value.isoformat(),
         )
+    else:
+        for encoder in CUSTOM_ENCODERS:
+            try:
+                rv = encoder(value)
+            except TypeError:
+                pass
+            else:
+                if not isinstance(rv, Item):
+                    raise _ConvertError(
+                        f"Custom encoder returned {type(rv)}, not a subclass of Item"
+                    )
+                return rv
 
-    raise ValueError(f"Invalid type {type(value)}")
-
-
-# This code is only valid for Python < 3.8, when @cached_property was introduced
-# it replaces chained @property and @lru_cache decorators
-def lazy_property(f):
-    return property(lru_cache(maxsize=None)(f))
+    raise _ConvertError(f"Invalid type {type(value)}")
 
 
 class StringType(Enum):
@@ -258,15 +238,8 @@ class StringType(Enum):
     # Multi Line Literal
     MLL = "'''"
 
-    def __init__(self, value):
-        self.is_basic = lru_cache(maxsize=None)(self._is_basic)
-        self.is_literal = lru_cache(maxsize=None)(self._is_literal)
-        self.is_singleline = lru_cache(maxsize=None)(self._is_singleline)
-        self.is_multiline = lru_cache(maxsize=None)(self._is_multiline)
-        self.toggle = lru_cache(maxsize=None)(self._toggle)
-
     @classmethod
-    def select(cls, literal=False, multiline=False) -> "StringType":
+    def select(cls, literal=False, multiline=False) -> StringType:
         return {
             (False, False): cls.SLB,
             (False, True): cls.MLB,
@@ -298,23 +271,23 @@ class StringType(Enum):
             StringType.MLL: (forbidden_in_literal | {"'''"}) - allowed_in_multiline,
         }[self]
 
-    @lazy_property
+    @property
     def unit(self) -> str:
         return self.value[0]
 
-    def _is_basic(self) -> bool:
+    def is_basic(self) -> bool:
         return self in {StringType.SLB, StringType.MLB}
 
-    def _is_literal(self) -> bool:
+    def is_literal(self) -> bool:
         return self in {StringType.SLL, StringType.MLL}
 
-    def _is_singleline(self) -> bool:
+    def is_singleline(self) -> bool:
         return self in {StringType.SLB, StringType.SLL}
 
-    def _is_multiline(self) -> bool:
+    def is_multiline(self) -> bool:
         return self in {StringType.MLB, StringType.MLL}
 
-    def _toggle(self) -> "StringType":
+    def toggle(self) -> StringType:
         return {
             StringType.SLB: StringType.MLB,
             StringType.MLB: StringType.SLB,
@@ -327,7 +300,6 @@ class BoolType(Enum):
     TRUE = "true"
     FALSE = "false"
 
-    @lru_cache(maxsize=None)  # noqa: B019
     def __bool__(self):
         return {BoolType.TRUE: True, BoolType.FALSE: False}[self]
 
@@ -338,32 +310,23 @@ class BoolType(Enum):
         return len(self.value)
 
 
+@dataclasses.dataclass
 class Trivia:
     """
     Trivia information (aka metadata).
     """
 
-    def __init__(
-        self,
-        indent: str = None,
-        comment_ws: str = None,
-        comment: str = None,
-        trail: str = None,
-    ) -> None:
-        # Whitespace before a value.
-        self.indent = indent or ""
-        # Whitespace after a value, but before a comment.
-        self.comment_ws = comment_ws or ""
-        # Comment, starting with # character, or empty string if no comment.
-        self.comment = comment or ""
-        # Trailing newline.
-        if trail is None:
-            trail = "\n"
+    # Whitespace before a value.
+    indent: str = ""
+    # Whitespace after a value, but before a comment.
+    comment_ws: str = ""
+    # Comment, starting with # character, or empty string if no comment.
+    comment: str = ""
+    # Trailing newline.
+    trail: str = "\n"
 
-        self.trail = trail
-
-    def copy(self) -> "Trivia":
-        return type(self)(self.indent, self.comment_ws, self.comment, self.trail)
+    def copy(self) -> Trivia:
+        return dataclasses.replace(self)
 
 
 class KeyType(Enum):
@@ -384,7 +347,7 @@ class Key(abc.ABC):
 
     sep: str
     _original: str
-    _keys: List["SingleKey"]
+    _keys: list[SingleKey]
     _dotted: bool
     key: str
 
@@ -400,10 +363,10 @@ class Key(abc.ABC):
         """If the key is followed by other keys"""
         return self._dotted
 
-    def __iter__(self) -> Iterator["SingleKey"]:
+    def __iter__(self) -> Iterator[SingleKey]:
         return iter(self._keys)
 
-    def concat(self, other: "Key") -> "DottedKey":
+    def concat(self, other: Key) -> DottedKey:
         """Concatenate keys into a dotted key"""
         keys = self._keys + other._keys
         return DottedKey(keys, sep=self.sep)
@@ -429,9 +392,9 @@ class SingleKey(Key):
     def __init__(
         self,
         k: str,
-        t: Optional[KeyType] = None,
-        sep: Optional[str] = None,
-        original: Optional[str] = None,
+        t: KeyType | None = None,
+        sep: str | None = None,
+        original: str | None = None,
     ) -> None:
         if t is None:
             if not k or any(
@@ -477,9 +440,9 @@ class SingleKey(Key):
 class DottedKey(Key):
     def __init__(
         self,
-        keys: Iterable[Key],
-        sep: Optional[str] = None,
-        original: Optional[str] = None,
+        keys: Iterable[SingleKey],
+        sep: str | None = None,
+        original: str | None = None,
     ) -> None:
         self._keys = list(keys)
         if original is None:
@@ -518,13 +481,17 @@ class Item:
         """The TOML representation"""
         raise NotImplementedError()
 
-    def unwrap(self):
+    @property
+    def value(self) -> Any:
+        return self
+
+    def unwrap(self) -> Any:
         """Returns as pure python object (ppo)"""
         raise NotImplementedError()
 
     # Helpers
 
-    def comment(self, comment: str) -> "Item":
+    def comment(self, comment: str) -> Item:
         """Attach a comment to this item"""
         if not comment.strip().startswith("#"):
             comment = "# " + comment
@@ -534,7 +501,7 @@ class Item:
 
         return self
 
-    def indent(self, indent: int) -> "Item":
+    def indent(self, indent: int) -> Item:
         """Indent this item with given number of spaces"""
         if self._trivia.indent.startswith("\n"):
             self._trivia.indent = "\n" + " " * indent
@@ -623,17 +590,17 @@ class Comment(Item):
         return f"{self._trivia.indent}{decode(self._trivia.comment)}"
 
 
-class Integer(int, Item):
+class Integer(Item, _CustomInt):
     """
     An integer literal.
     """
 
-    def __new__(cls, value: int, trivia: Trivia, raw: str) -> "Integer":
-        return super().__new__(cls, value)
+    def __new__(cls, value: int, trivia: Trivia, raw: str) -> Integer:
+        return int.__new__(cls, value)
 
-    def __init__(self, _: int, trivia: Trivia, raw: str) -> None:
+    def __init__(self, value: int, trivia: Trivia, raw: str) -> None:
         super().__init__(trivia)
-
+        self._original = value
         self._raw = raw
         self._sign = False
 
@@ -641,7 +608,12 @@ class Integer(int, Item):
             self._sign = True
 
     def unwrap(self) -> int:
-        return int(self)
+        return self._original
+
+    __int__ = unwrap
+
+    def __hash__(self) -> int:
+        return hash(self.unwrap())
 
     @property
     def discriminant(self) -> int:
@@ -655,30 +627,6 @@ class Integer(int, Item):
     def as_string(self) -> str:
         return self._raw
 
-    def __add__(self, other):
-        return self._new(int(self._raw) + other)
-
-    def __radd__(self, other):
-        result = super().__radd__(other)
-
-        if isinstance(other, Integer):
-            return self._new(result)
-
-        return result
-
-    def __sub__(self, other):
-        result = super().__sub__(other)
-
-        return self._new(result)
-
-    def __rsub__(self, other):
-        result = super().__rsub__(other)
-
-        if isinstance(other, Integer):
-            return self._new(result)
-
-        return result
-
     def _new(self, result):
         raw = str(result)
         if self._sign:
@@ -690,18 +638,53 @@ class Integer(int, Item):
     def _getstate(self, protocol=3):
         return int(self), self._trivia, self._raw
 
+    # int methods
+    __abs__ = wrap_method(int.__abs__)
+    __add__ = wrap_method(int.__add__)
+    __and__ = wrap_method(int.__and__)
+    __ceil__ = wrap_method(int.__ceil__)
+    __eq__ = int.__eq__
+    __floor__ = wrap_method(int.__floor__)
+    __floordiv__ = wrap_method(int.__floordiv__)
+    __invert__ = wrap_method(int.__invert__)
+    __le__ = int.__le__
+    __lshift__ = wrap_method(int.__lshift__)
+    __lt__ = int.__lt__
+    __mod__ = wrap_method(int.__mod__)
+    __mul__ = wrap_method(int.__mul__)
+    __neg__ = wrap_method(int.__neg__)
+    __or__ = wrap_method(int.__or__)
+    __pos__ = wrap_method(int.__pos__)
+    __pow__ = wrap_method(int.__pow__)
+    __radd__ = wrap_method(int.__radd__)
+    __rand__ = wrap_method(int.__rand__)
+    __rfloordiv__ = wrap_method(int.__rfloordiv__)
+    __rlshift__ = wrap_method(int.__rlshift__)
+    __rmod__ = wrap_method(int.__rmod__)
+    __rmul__ = wrap_method(int.__rmul__)
+    __ror__ = wrap_method(int.__ror__)
+    __round__ = wrap_method(int.__round__)
+    __rpow__ = wrap_method(int.__rpow__)
+    __rrshift__ = wrap_method(int.__rrshift__)
+    __rshift__ = wrap_method(int.__rshift__)
+    __rtruediv__ = wrap_method(int.__rtruediv__)
+    __rxor__ = wrap_method(int.__rxor__)
+    __truediv__ = wrap_method(int.__truediv__)
+    __trunc__ = wrap_method(int.__trunc__)
+    __xor__ = wrap_method(int.__xor__)
 
-class Float(float, Item):
+
+class Float(Item, _CustomFloat):
     """
     A float literal.
     """
 
-    def __new__(cls, value: float, trivia: Trivia, raw: str) -> Integer:
-        return super().__new__(cls, value)
+    def __new__(cls, value: float, trivia: Trivia, raw: str) -> Float:
+        return float.__new__(cls, value)
 
-    def __init__(self, _: float, trivia: Trivia, raw: str) -> None:
+    def __init__(self, value: float, trivia: Trivia, raw: str) -> None:
         super().__init__(trivia)
-
+        self._original = value
         self._raw = raw
         self._sign = False
 
@@ -709,7 +692,12 @@ class Float(float, Item):
             self._sign = True
 
     def unwrap(self) -> float:
-        return float(self)
+        return self._original
+
+    __float__ = unwrap
+
+    def __hash__(self) -> int:
+        return hash(self.unwrap())
 
     @property
     def discriminant(self) -> int:
@@ -723,32 +711,6 @@ class Float(float, Item):
     def as_string(self) -> str:
         return self._raw
 
-    def __add__(self, other):
-        result = super().__add__(other)
-
-        return self._new(result)
-
-    def __radd__(self, other):
-        result = super().__radd__(other)
-
-        if isinstance(other, Float):
-            return self._new(result)
-
-        return result
-
-    def __sub__(self, other):
-        result = super().__sub__(other)
-
-        return self._new(result)
-
-    def __rsub__(self, other):
-        result = super().__rsub__(other)
-
-        if isinstance(other, Float):
-            return self._new(result)
-
-        return result
-
     def _new(self, result):
         raw = str(result)
 
@@ -761,6 +723,35 @@ class Float(float, Item):
     def _getstate(self, protocol=3):
         return float(self), self._trivia, self._raw
 
+    # float methods
+    __abs__ = wrap_method(float.__abs__)
+    __add__ = wrap_method(float.__add__)
+    __eq__ = float.__eq__
+    __floordiv__ = wrap_method(float.__floordiv__)
+    __le__ = float.__le__
+    __lt__ = float.__lt__
+    __mod__ = wrap_method(float.__mod__)
+    __mul__ = wrap_method(float.__mul__)
+    __neg__ = wrap_method(float.__neg__)
+    __pos__ = wrap_method(float.__pos__)
+    __pow__ = wrap_method(float.__pow__)
+    __radd__ = wrap_method(float.__radd__)
+    __rfloordiv__ = wrap_method(float.__rfloordiv__)
+    __rmod__ = wrap_method(float.__rmod__)
+    __rmul__ = wrap_method(float.__rmul__)
+    __round__ = wrap_method(float.__round__)
+    __rpow__ = wrap_method(float.__rpow__)
+    __rtruediv__ = wrap_method(float.__rtruediv__)
+    __truediv__ = wrap_method(float.__truediv__)
+    __trunc__ = float.__trunc__
+
+    if sys.version_info >= (3, 9):
+        __ceil__ = float.__ceil__
+        __floor__ = float.__floor__
+    else:
+        __ceil__ = math.ceil
+        __floor__ = math.floor
+
 
 class Bool(Item):
     """
@@ -822,7 +813,7 @@ class DateTime(Item, datetime):
         minute: int,
         second: int,
         microsecond: int,
-        tzinfo: Optional[tzinfo],
+        tzinfo: tzinfo | None,
         *_: Any,
         **kwargs: Any,
     ) -> datetime:
@@ -848,9 +839,9 @@ class DateTime(Item, datetime):
         minute: int,
         second: int,
         microsecond: int,
-        tzinfo: Optional[tzinfo],
-        trivia: Optional[Trivia] = None,
-        raw: Optional[str] = None,
+        tzinfo: tzinfo | None,
+        trivia: Trivia | None = None,
+        raw: str | None = None,
         **kwargs: Any,
     ) -> None:
         super().__init__(trivia or Trivia())
@@ -929,7 +920,7 @@ class DateTime(Item, datetime):
             return result
         return self._new(result)
 
-    def _new(self, result) -> "DateTime":
+    def _new(self, result) -> DateTime:
         raw = result.isoformat()
 
         return DateTime(
@@ -1032,7 +1023,7 @@ class Time(Item, time):
         minute: int,
         second: int,
         microsecond: int,
-        tzinfo: Optional[tzinfo],
+        tzinfo: tzinfo | None,
         *_: Any,
     ) -> time:
         return time.__new__(cls, hour, minute, second, microsecond, tzinfo)
@@ -1043,7 +1034,7 @@ class Time(Item, time):
         minute: int,
         second: int,
         microsecond: int,
-        tzinfo: Optional[tzinfo],
+        tzinfo: tzinfo | None,
         trivia: Trivia,
         raw: str,
     ) -> None:
@@ -1051,7 +1042,7 @@ class Time(Item, time):
 
         self._raw = raw
 
-    def unwrap(self) -> datetime:
+    def unwrap(self) -> time:
         (hour, minute, second, microsecond, tzinfo, _, _) = self._getstate()
         return time(hour, minute, second, microsecond, tzinfo)
 
@@ -1099,10 +1090,10 @@ class _ArrayItemGroup:
 
     def __init__(
         self,
-        value: Optional[Item] = None,
-        indent: Optional[Whitespace] = None,
-        comma: Optional[Whitespace] = None,
-        comment: Optional[Comment] = None,
+        value: Item | None = None,
+        indent: Whitespace | None = None,
+        comma: Whitespace | None = None,
+        comment: Comment | None = None,
     ) -> None:
         self.value = value
         self.indent = indent
@@ -1134,19 +1125,19 @@ class Array(Item, _CustomList):
     """
 
     def __init__(
-        self, value: List[Item], trivia: Trivia, multiline: bool = False
+        self, value: list[Item], trivia: Trivia, multiline: bool = False
     ) -> None:
         super().__init__(trivia)
         list.__init__(
             self,
-            [v.value for v in value if not isinstance(v, (Whitespace, Comment, Null))],
+            [v for v in value if not isinstance(v, (Whitespace, Comment, Null))],
         )
-        self._index_map: Dict[int, int] = {}
+        self._index_map: dict[int, int] = {}
         self._value = self._group_values(value)
         self._multiline = multiline
         self._reindex()
 
-    def _group_values(self, value: List[Item]) -> List[_ArrayItemGroup]:
+    def _group_values(self, value: list[Item]) -> list[_ArrayItemGroup]:
         """Group the values into (indent, value, comma, comment) tuples"""
         groups = []
         this_group = _ArrayItemGroup()
@@ -1172,10 +1163,10 @@ class Array(Item, _CustomList):
         groups.append(this_group)
         return [group for group in groups if group]
 
-    def unwrap(self) -> str:
+    def unwrap(self) -> list[Any]:
         unwrapped = []
         for v in self:
-            if isinstance(v, Item):
+            if hasattr(v, "unwrap"):
                 unwrapped.append(v.unwrap())
             else:
                 unwrapped.append(v)
@@ -1193,7 +1184,7 @@ class Array(Item, _CustomList):
         for v in self._value:
             yield from v
 
-    def multiline(self, multiline: bool) -> "Array":
+    def multiline(self, multiline: bool) -> Array:
         """Change the array to display in multiline or not.
 
         :Example:
@@ -1244,7 +1235,7 @@ class Array(Item, _CustomList):
         self,
         *items: Any,
         indent: str = "    ",
-        comment: Optional[str] = None,
+        comment: str | None = None,
         add_comma: bool = True,
         newline: bool = True,
     ) -> None:
@@ -1264,7 +1255,7 @@ class Array(Item, _CustomList):
             4, 5, 6,
         ]
         """
-        new_values: List[Item] = []
+        new_values: list[Item] = []
         first_indent = f"\n{indent}" if newline else indent
         if first_indent:
             new_values.append(Whitespace(first_indent))
@@ -1325,12 +1316,15 @@ class Array(Item, _CustomList):
     def __len__(self) -> int:
         return list.__len__(self)
 
-    def __getitem__(self, key: Union[int, slice]) -> Any:
-        return list.__getitem__(self, key)
+    def __getitem__(self, key: int | slice) -> Any:
+        rv = cast(Item, list.__getitem__(self, key))
+        if rv.is_boolean():
+            return bool(rv)
+        return rv
 
-    def __setitem__(self, key: Union[int, slice], value: Any) -> Any:
+    def __setitem__(self, key: int | slice, value: Any) -> Any:
         it = item(value, _parent=self)
-        list.__setitem__(self, key, it.value)
+        list.__setitem__(self, key, it)
         if isinstance(key, slice):
             raise ValueError("slice assignment is not supported")
         if key < 0:
@@ -1341,7 +1335,7 @@ class Array(Item, _CustomList):
         it = item(value, _parent=self)
         length = len(self)
         if not isinstance(it, (Comment, Whitespace)):
-            list.insert(self, pos, it.value)
+            list.insert(self, pos, it)
         if pos < 0:
             pos += length
             if pos < 0:
@@ -1364,8 +1358,8 @@ class Array(Item, _CustomList):
                     and "\n" in self._value[idx].indent.s
                 ):
                     default_indent = "\n    "
-        indent: Optional[Item] = None
-        comma: Optional[Item] = Whitespace(",") if pos < length else None
+        indent: Item | None = None
+        comma: Item | None = Whitespace(",") if pos < length else None
         if idx < len(self._value) and not self._value[idx].is_whitespace():
             # Prefer to copy the indentation from the item after
             indent = self._value[idx].indent
@@ -1387,7 +1381,7 @@ class Array(Item, _CustomList):
         self._value.insert(idx, new_item)
         self._reindex()
 
-    def __delitem__(self, key: Union[int, slice]):
+    def __delitem__(self, key: int | slice):
         length = len(self)
         list.__delitem__(self, key)
 
@@ -1420,20 +1414,14 @@ class Array(Item, _CustomList):
 
         self._reindex()
 
-    def __str__(self):
-        return str([v.value.value for v in self._iter_items() if v.value is not None])
-
     def _getstate(self, protocol=3):
         return list(self._iter_items()), self._trivia, self._multiline
 
 
-AT = TypeVar("AT", bound="AbstractTable")
-
-
 class AbstractTable(Item, _CustomDict):
     """Common behaviour of both :class:`Table` and :class:`InlineTable`"""
 
-    def __init__(self, value: "container.Container", trivia: Trivia):
+    def __init__(self, value: container.Container, trivia: Trivia):
         Item.__init__(self, trivia)
 
         self._value = value
@@ -1442,38 +1430,38 @@ class AbstractTable(Item, _CustomDict):
             if k is not None:
                 dict.__setitem__(self, k.key, v)
 
-    def unwrap(self):
+    def unwrap(self) -> dict[str, Any]:
         unwrapped = {}
         for k, v in self.items():
             if isinstance(k, Key):
                 k = k.key
-            if isinstance(v, Item):
+            if hasattr(v, "unwrap"):
                 v = v.unwrap()
             unwrapped[k] = v
 
         return unwrapped
 
     @property
-    def value(self) -> "container.Container":
+    def value(self) -> container.Container:
         return self._value
 
     @overload
-    def append(self: AT, key: None, value: Union[Comment, Whitespace]) -> AT:
+    def append(self: AT, key: None, value: Comment | Whitespace) -> AT:
         ...
 
     @overload
-    def append(self: AT, key: Union[Key, str], value: Any) -> AT:
+    def append(self: AT, key: Key | str, value: Any) -> AT:
         ...
 
     def append(self, key, value):
         raise NotImplementedError
 
     @overload
-    def add(self: AT, value: Union[Comment, Whitespace]) -> AT:
+    def add(self: AT, key: Comment | Whitespace) -> AT:
         ...
 
     @overload
-    def add(self: AT, key: Union[Key, str], value: Any) -> AT:
+    def add(self: AT, key: Key | str, value: Any = ...) -> AT:
         ...
 
     def add(self, key, value=None):
@@ -1486,7 +1474,7 @@ class AbstractTable(Item, _CustomDict):
 
         return self.append(key, value)
 
-    def remove(self: AT, key: Union[Key, str]) -> AT:
+    def remove(self: AT, key: Key | str) -> AT:
         self._value.remove(key)
 
         if isinstance(key, Key):
@@ -1497,7 +1485,7 @@ class AbstractTable(Item, _CustomDict):
 
         return self
 
-    def setdefault(self, key: Union[Key, str], default: Any) -> Any:
+    def setdefault(self, key: Key | str, default: Any) -> Any:
         super().setdefault(key, default)
         return self[key]
 
@@ -1516,15 +1504,15 @@ class AbstractTable(Item, _CustomDict):
     def __len__(self) -> int:
         return len(self._value)
 
-    def __delitem__(self, key: Union[Key, str]) -> None:
+    def __delitem__(self, key: Key | str) -> None:
         self.remove(key)
 
-    def __getitem__(self, key: Union[Key, str]) -> Item:
+    def __getitem__(self, key: Key | str) -> Item:
         return cast(Item, self._value[key])
 
-    def __setitem__(self, key: Union[Key, str], value: Any) -> None:
+    def __setitem__(self, key: Key | str, value: Any) -> None:
         if not isinstance(value, Item):
-            value = item(value)
+            value = item(value, _parent=self)
 
         is_replace = key in self
         self._value[key] = value
@@ -1555,12 +1543,12 @@ class Table(AbstractTable):
 
     def __init__(
         self,
-        value: "container.Container",
+        value: container.Container,
         trivia: Trivia,
         is_aot_element: bool,
-        is_super_table: bool = False,
-        name: Optional[str] = None,
-        display_name: Optional[str] = None,
+        is_super_table: bool | None = None,
+        name: str | None = None,
+        display_name: str | None = None,
     ) -> None:
         super().__init__(value, trivia)
 
@@ -1573,7 +1561,7 @@ class Table(AbstractTable):
     def discriminant(self) -> int:
         return 9
 
-    def __copy__(self) -> "Table":
+    def __copy__(self) -> Table:
         return type(self)(
             self._value.copy(),
             self._trivia.copy(),
@@ -1583,12 +1571,12 @@ class Table(AbstractTable):
             self.display_name,
         )
 
-    def append(self, key, _item):
+    def append(self, key: Key | str | None, _item: Any) -> Table:
         """
         Appends a (key, item) to the table.
         """
         if not isinstance(_item, Item):
-            _item = item(_item)
+            _item = item(_item, _parent=self)
 
         self._value.append(key, _item)
 
@@ -1614,7 +1602,7 @@ class Table(AbstractTable):
 
         return self
 
-    def raw_append(self, key: Union[Key, str], _item: Any) -> "Table":
+    def raw_append(self, key: Key | str | None, _item: Any) -> Table:
         """Similar to :meth:`append` but does not copy indentation."""
         if not isinstance(_item, Item):
             _item = item(_item)
@@ -1637,14 +1625,20 @@ class Table(AbstractTable):
     def is_super_table(self) -> bool:
         """A super table is the intermediate parent of a nested table as in [a.b.c].
         If true, it won't appear in the TOML representation."""
-        return self._is_super_table
+        if self._is_super_table is not None:
+            return self._is_super_table
+        # If the table has only one child and that child is a table, then it is a super table.
+        if len(self) != 1:
+            return False
+        only_child = next(iter(self.values()))
+        return isinstance(only_child, (Table, AoT))
 
     def as_string(self) -> str:
         return self._value.as_string()
 
     # Helpers
 
-    def indent(self, indent: int) -> "Table":
+    def indent(self, indent: int) -> Table:
         """Indent the table with given number of spaces."""
         super().indent(indent)
 
@@ -1684,7 +1678,7 @@ class InlineTable(AbstractTable):
     """
 
     def __init__(
-        self, value: "container.Container", trivia: Trivia, new: bool = False
+        self, value: container.Container, trivia: Trivia, new: bool = False
     ) -> None:
         super().__init__(value, trivia)
 
@@ -1694,12 +1688,12 @@ class InlineTable(AbstractTable):
     def discriminant(self) -> int:
         return 10
 
-    def append(self, key, _item):
+    def append(self, key: Key | str | None, _item: Any) -> InlineTable:
         """
         Appends a (key, item) to the table.
         """
         if not isinstance(_item, Item):
-            _item = item(_item)
+            _item = item(_item, _parent=self)
 
         if not isinstance(_item, (Whitespace, Comment)):
             if not _item.trivia.indent and len(self._value) > 0 and not self._new:
@@ -1719,6 +1713,14 @@ class InlineTable(AbstractTable):
 
     def as_string(self) -> str:
         buf = "{"
+        last_item_idx = next(
+            (
+                i
+                for i in range(len(self._value.body) - 1, -1, -1)
+                if self._value.body[i][0] is not None
+            ),
+            None,
+        )
         for i, (k, v) in enumerate(self._value.body):
             if k is None:
                 if i == len(self._value.body) - 1:
@@ -1741,7 +1743,7 @@ class InlineTable(AbstractTable):
                 f"{v_trivia_trail}"
             )
 
-            if i != len(self._value.body) - 1:
+            if last_item_idx is not None and i < last_item_idx:
                 buf += ","
                 if self._new:
                     buf += " "
@@ -1750,12 +1752,12 @@ class InlineTable(AbstractTable):
 
         return buf
 
-    def __setitem__(self, key: Union[Key, str], value: Any) -> None:
+    def __setitem__(self, key: Key | str, value: Any) -> None:
         if hasattr(value, "trivia") and value.trivia.comment:
             value.trivia.comment = ""
         super().__setitem__(key, value)
 
-    def __copy__(self) -> "InlineTable":
+    def __copy__(self) -> InlineTable:
         return type(self)(self._value.copy(), self._trivia.copy(), self._new)
 
     def _getstate(self, protocol: int = 3) -> tuple:
@@ -1790,24 +1792,22 @@ class String(str, Item):
     def as_string(self) -> str:
         return f"{self._t.value}{decode(self._original)}{self._t.value}"
 
-    def __add__(self, other):
+    def __add__(self: ItemT, other: str) -> ItemT:
+        if not isinstance(other, str):
+            return NotImplemented
         result = super().__add__(other)
+        original = self._original + getattr(other, "_original", other)
 
-        return self._new(result)
+        return self._new(result, original)
 
-    def __sub__(self, other):
-        result = super().__sub__(other)
-
-        return self._new(result)
-
-    def _new(self, result):
-        return String(self._t, result, result, self._trivia)
+    def _new(self, result: str, original: str) -> String:
+        return String(self._t, result, original, self._trivia)
 
     def _getstate(self, protocol=3):
         return self._t, str(self), self._original, self._trivia
 
     @classmethod
-    def from_raw(cls, value: str, type_=StringType.SLB, escape=True) -> "String":
+    def from_raw(cls, value: str, type_=StringType.SLB, escape=True) -> String:
         value = decode(value)
 
         invalid = type_.invalid_sequences
@@ -1826,10 +1826,10 @@ class AoT(Item, _CustomList):
     """
 
     def __init__(
-        self, body: List[Table], name: Optional[str] = None, parsed: bool = False
+        self, body: list[Table], name: str | None = None, parsed: bool = False
     ) -> None:
         self.name = name
-        self._body: List[Table] = []
+        self._body: list[Table] = []
         self._parsed = parsed
 
         super().__init__(Trivia(trail=""))
@@ -1837,17 +1837,17 @@ class AoT(Item, _CustomList):
         for table in body:
             self.append(table)
 
-    def unwrap(self) -> str:
+    def unwrap(self) -> list[dict[str, Any]]:
         unwrapped = []
         for t in self._body:
-            if isinstance(t, Item):
+            if hasattr(t, "unwrap"):
                 unwrapped.append(t.unwrap())
             else:
                 unwrapped.append(t)
         return unwrapped
 
     @property
-    def body(self) -> List[Table]:
+    def body(self) -> list[Table]:
         return self._body
 
     @property
@@ -1855,14 +1855,14 @@ class AoT(Item, _CustomList):
         return 12
 
     @property
-    def value(self) -> List[Dict[Any, Any]]:
+    def value(self) -> list[dict[Any, Any]]:
         return [v.value for v in self._body]
 
     def __len__(self) -> int:
         return len(self._body)
 
     @overload
-    def __getitem__(self, key: slice) -> List[Table]:
+    def __getitem__(self, key: slice) -> list[Table]:
         ...
 
     @overload
@@ -1872,10 +1872,10 @@ class AoT(Item, _CustomList):
     def __getitem__(self, key):
         return self._body[key]
 
-    def __setitem__(self, key: Union[slice, int], value: Any) -> None:
+    def __setitem__(self, key: slice | int, value: Any) -> None:
         raise NotImplementedError
 
-    def __delitem__(self, key: Union[slice, int]) -> None:
+    def __delitem__(self, key: slice | int) -> None:
         del self._body[key]
         list.__delitem__(self, key)
 
@@ -1937,7 +1937,7 @@ class Null(Item):
     def __init__(self) -> None:
         pass
 
-    def unwrap(self) -> str:
+    def unwrap(self) -> None:
         return None
 
     @property
